#include <mem.h>
#include <stdio.h>

typedef unsigned char byte, *pbyte;
typedef unsigned int uint, *puint;


#define RAW_BYTES_CNT     0x80
#define RAW_INTS_CNT      (RAW_BYTES_CNT/4)      // 0x20
#define RAW_BITS_CNT      (RAW_BYTES_CNT*8)      // 0x400


void Decrypt (puint data);
void Transform (puint data1, puint data2, puint key);

// -     (1024 , -)
//    ,     
//     gcc,    AT&T (  -  )
//      (CF)
bool RCL (puint dest,  uint cnt, int len = RAW_INTS_CNT);
bool ADC (puint dest, puint src, int len = RAW_INTS_CNT);
bool SBB (puint dest, puint src, int len = RAW_INTS_CNT);



// ,  
uint decKey[] = {
  0x5207F766,0x01802B20,0xD521BAA7,0x413DF910,
  0x98A2CFB5,0x0BCE3D29,0x7822AC4A,0x43EDC1F8,
  0x181313F1,0x0A3681CD,0x7CA1D69C,0xE1731780,
  0x8B4894D9,0x08431C6E,0x731496AF,0xF3C86456,
  0x14DB3E5F,0xF4A5E921,0x6B87751A,0xCF19E5A0,
  0x61F748D1,0x3DD28192,0xD71154E3,0x609781CB,
  0x5E94F797,0x1308C784,0x581D3FEF,0x71E457EF,
  0x35A9BC10,0x32E719F4,0x7F14530F,0x1A137F29
};



// decipher sequence
//         
//  ,   
byte decSeq[] = {0x71,0x4D,0xF0};




int main () {

  byte data[RAW_BYTES_CNT];
  int bytesWritten;

  FILE* hFile = fopen ("rawdata.bin", "rb");

  if (hFile) {
    memset (data, 0, RAW_BYTES_CNT);
    uint bytesRead = fread (data, sizeof(byte), RAW_BYTES_CNT, hFile);
    fclose (hFile);

    if (bytesRead == RAW_BYTES_CNT) {

      Decrypt ((puint)data);

      hFile = fopen ("decrypted.bin", "wb");

      if (hFile) {
        bytesWritten = fwrite (data, sizeof(byte), RAW_BYTES_CNT, hFile);
        fclose (hFile);

        if (bytesWritten != RAW_BYTES_CNT)
          printf ("fwrite (\"decrypted.bin\") failed\n");
      }
      else printf ("fopen (\"decrypted.bin\") failed\n");
    }
    else printf ("fread (\"rawdata.bin\") failed\n");
  }
  else printf ("fopen (\"rawdata.bin\") failed\n");

  return 0;
}





//        .   - 
void Decrypt (puint data) {

  uint var1[RAW_INTS_CNT];
  uint var2[RAW_INTS_CNT];

  memset (var1, 0, RAW_BYTES_CNT);
  var1[0] = 1;

  //     decSeq
  // i -   , j -    
  for (int i = 0; i < sizeof(decSeq); i++)
    for (int j = 7; j >= 0; j--) {

      memcpy (var2, var1, RAW_BYTES_CNT);
      Transform (var1, var2, decKey);

      //        decSeq,
      //     
      if (decSeq[i] & 1 << j)
        Transform (var1, data, decKey);
    }

  memcpy (data, var1, RAW_BYTES_CNT);
}





void Transform (puint data1, puint data2, puint key) {

  uint var[RAW_INTS_CNT];

  memset (var, 0, RAW_BYTES_CNT);

  //     var
  for (int i = 0; i < RAW_BITS_CNT; i++) {
    RCL (var, 1);
    if (SBB (var, key))  ADC (var, key);
    if (RCL (data1, 1))  ADC (var, data2);
    if (SBB (var, key))  ADC (var, key);
  }

  memcpy (data1, var, RAW_BYTES_CNT);
}





bool RCL (puint dest, uint cnt, int len) {

  bool carry;

/*
  clc
@@:
  lodsd
  rcl eax, cl
  stosd
  dec ebx
    jg @b
  setc al
  movzx eax, al
*/

  asm (
    "  clc                              ;"
    "rclloop:                            "
    "  lodsl                            ;"
    "  rcl %%cl, %%eax                  ;"
    "  stosl                            ;"
    "  decl %%ebx                       ;"
    "    jg rclloop                     ;"
    "  setc %%al                        ;"
    "  movzx %%al, %%eax                ;"
  : "=a" (carry)
  : "c" (cnt), "b" (len), "S" (dest), "D" (dest)
  );

  return carry;
}





bool ADC (puint dest, puint src, int len) {

  bool carry;

/*
  clc
@@:
  lodsd
  adc [edi], eax
  lea edi, [edi+4]  // lea  add,   
    loop @b
  setc al
  movzx eax, al
*/

  asm (
    "  clc                              ;"
    "adcloop:                            "
    "  lodsl                            ;"
    "  adcl %%eax, (%%edi)              ;"
    "  leal 4(%%edi),%%edi              ;"
    "    loop adcloop                   ;"
    "  setc %%al                        ;"
    "  movzx %%al, %%eax                ;"
  : "=a" (carry)
  : "c" (len), "S" (src), "D" (dest)
  );

  return carry;
}





bool SBB (puint dest, puint src, int len) {

  bool carry;

/*
  clc
@@:
  lodsd
  sbb [edi], eax
  lea edi, [edi+4]  // lea  add,   
    loop @b
  setc al
  movzx eax, al
*/

  asm (
    "  clc                              ;"
    "sbbloop:                            "
    "  lodsl                            ;"
    "  sbbl %%eax, (%%edi)              ;"
    "  leal 4(%%edi),%%edi              ;"
    "    loop sbbloop                   ;"
    "  setc %%al                        ;"
    "  movzx %%al, %%eax                ;"
  : "=a" (carry)
  : "c" (len), "S" (src), "D" (dest)
  );

  return carry;
}




